#include "TQuaternion.h"
#include "Riostream.h"
   
void TestQuaternion() {

    //gSystem->Load("libPhysics");


    // test constructors :
    TQuaternion q0;
    TQuaternion q1(1);
    TQuaternion q2(0.5,0.1,0.2,0.3);
    TQuaternion q3(TVector3(0.1,0.2,0.3),0.5);
    double ratio;
    double ArrD[4] = { 0.1, 0.2, 0.3, 0.5};
    float ArrF[4] = { 0.1, 0.2, 0.3, 0.7};
    TQuaternion q4(ArrD);
    TQuaternion q5(ArrF);
    TQuaternion q6(q5);

    cout << endl<< "============ TESTING CONSTRUCTOR  " << endl;
    cout << "q0 : " << endl;
    q0.Print(); cout << endl;
    cout << "q1(1) : "<< endl;
    q1.Print(); cout << endl;
    cout << "q2(0.5,0.1,0.2,0.3) : "<< endl;
    q2.Print(); cout << endl;
    cout << "q3(TVector3(0.1,0.2,0.3),0.5) : "<< endl;
    q3.Print(); cout << endl;
    cout << "q4(double[4]= { 0.1, 0.2, 0.3, 0.5}) : "<< endl;
    q4.Print(); cout << endl;
    cout << "q5(float[4]= { 0.1, 0.2, 0.3, 0.7}) : "<< endl;
    q5.Print();cout << endl;
    cout << "q6(q5) : "<< endl;
    q6.Print(); cout << endl;

    int i;
   cout << endl<< "============ TESTING operator (int) & [int]  " << endl;
   for (i=0; i<4; i++) {
        cout << "q5("<< i<< ")=" << q5(i) << endl;
        cout << "q5["<< i<< "]=" << q5[i] << endl;
   }

  cout << endl<< "============ TESTING Get / SetRXYZ  " << endl;
  q0.SetRXYZ(0.2,0.3,0.4,0.5);
   cout << "q0.SetRXYZ(0.2,0.3,0.4,0.5) => " << endl;
   q0.Print(); cout << endl;

   q0.GetRXYZ(ArrD);
   q0.GetRXYZ(ArrF);
   for (i=0; i<4; i++) {
        cout << "ArrD <- q5 : [" << i << "] =" << ArrD[i] << endl;
        cout << "ArrF <- q5 : [" << i << "] =" << ArrF[i] << endl;
   }

  //--------------- real to Quaternion
   cout << endl << "============ TESTING real to Quaternion algebra  " << endl;

   q0 = 0.5;
   cout << "q0 = 0.5 " << endl;
   q0.Print();

   if (q0==0.5) cout << "OK ! "<< endl; else cout << " NOT OK ! " << endl;
   if (q0!=0.49) cout << "OK ! "<< endl; else cout << " NOT OK ! " << endl;


   cout << "q5 " << endl;
   q5.Print();
   cout << " q6 (= q5 initial) " << endl;
   q6.Print();

   q5+=0.5;
   cout << "q5+=0.5 " << endl;
   q5.Print();
   TQuaternion q7 = q6 + 0.5;
   cout << "q7 = q6 + 0.5 " << endl;
   q7.Print();
   if (q5==q7) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
   TQuaternion q97 = 0.5 + q6;
   cout << "q97 = 0.5 + q6 " << endl;
   q97.Print();
   if (q5==q97) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   q5-=0.3;
   cout << "q5-=0.3 " << endl;
   q5.Print();
   TQuaternion q8 = q7 - 0.3;
   cout << "q8 = q7 - 0.3 " << endl;
   q8.Print();
   if (q5==q8) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
   TQuaternion q98 = 0.3 -q7;
   q98 = -q98;
   cout << "q98 = -(0.3 - q7)" << endl;
   q98.Print();
   if (q5==q98) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   q5*=0.7;
   cout << "q5*=0.7 " << endl;
   q5.Print();
   TQuaternion q9 = q8 * 0.7;
   cout << "q9 = q8 * 0.7 " << endl;
   q9.Print();
   if (q5==q9) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
   TQuaternion q99 = 0.7 * q8;
   cout << "q99 = 0.7 * q8 " << endl;
   q99.Print();
   if (q5==q99) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   q5/=1.5;
   cout << "q5/=1.5 " << endl;
   q5.Print();
   TQuaternion q10 = q9 / 1.5;
   cout << "q10 = q9 / 1.5 " << endl;
   q10.Print();
   if (q5==q10) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
   TQuaternion q100 = 1.5 / q9;
   q100 = q100.Invert();
   cout << "q100 = 1/ (1.5 / q9) " << endl;
   q100.Print();
   if (q5==q100) cout << "OK !" << endl;
   else {
      cout << " ! NOT OK ! " << endl;
      cout << "DIFF = " << endl;
      (q5-q100).Print();
	  ratio = (q5-q100).Norm()/q100.Norm();
	  if (ratio < 1e-10) cout << "=> OK : ratio (1/ (1.5 / q9) - q9 / 1.5 ) / (q9 / 1.5) " << ratio << " infinitesimal => numericaly acceptable " << endl;
   }

  //--------------- vector to Quaternion
   cout << endl << "============ TESTING vector to Quaternion algebra  " << endl;
  TVector3 v(0.4,0.5,0.6);
  TVector3 W(0.4,0.5,0.59);

   q0 = v;
   cout << "q0 = v(0.4,0.5,0.6) " << endl;
   q0.Print();

   if (q0==v) cout << "OK ! "<< endl; else cout << " NOT OK ! " << endl;
   if (q0!=W) cout << "OK ! "<< endl; else cout << " NOT OK ! " << endl;

   cout << "q5 " << endl;
   q5.Print();
   q6 = q5;
   cout << " q6 (= q5 initial) " << endl;
   q6.Print();
   if (q5==q6) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   q5 += TVector3(0.1,0.2,0.3);
   cout << "q5+=TVector3(0.1,0.2,0.3) " << endl;
   q5.Print();
   TQuaternion q17 = q6 + TVector3(0.1,0.2,0.3);
   cout << "q17 = q6 + TVector3(0.1,0.2,0.3) " << endl;
   q17.Print();
   if (q5==q17) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
   TQuaternion q57 = TVector3(0.1,0.2,0.3) + q6;
   cout << "q57 = TVector3(0.1,0.2,0.3) + q6 " << endl;
   q57.Print();
   if (q5==q57) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   q5-=TVector3(0.5,0.6,0.7);
   cout << "q5-=TVector3(0.5,0.6,0.7) " << endl;
   q5.Print();
   TQuaternion q18 = q17 - TVector3(0.5,0.6,0.7);
   cout << "q18 = q17 - TVector3(0.5,0.6,0.7) " << endl;
   q18.Print();
   if (q5==q18) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
   TQuaternion q58 = TVector3(0.5,0.6,0.7) - q17;
   q58 = -q58;
   cout << "q58 = - (TVector3(0.5,0.6,0.7) - q17)" << endl;
   q58.Print();
   if (q5==q58) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   TVector3 V(0.6,0.7,0.8);
   cout << " Norm q5 " << q5.Norm() << " Norm V(0.6,0.7,0.8) " << V.Mag() << " prod= " << q5.Norm() * V.Mag() << endl;
   q5*=V;
   cout << "q5*=V " << endl;
   q5.Print();
   cout << "norm = " << q5.Norm() << endl;
   TQuaternion q19 = q18 * V;
   cout << "q19 = q18 * V " << endl;
   q19.Print();
   if (q5==q19) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
    // RMQ : non commutatif

   TVector3 V2(3,2,1);
   cout << " Norm q5 " << q5.Norm() << " Norm V2(3,2,1) " << V2.Mag() << " div= " << q5.Norm() / V2.Mag() << endl;
   q5/=V2;
   cout << "q5/=V2 " << endl;
   q5.Print();
   cout << "norm = " << q5.Norm() << endl;
   TQuaternion q20 = q19 / V2;
   cout << "q20 = q19 / V2 " << endl;
   q20.Print();
   if (q5==q20) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
    // RMQ : non commutatif

   // compare vector and quaternion multiplications
   cout << endl << "============ TESTING vector and quaternion multiplications  " << endl;

    TQuaternion QS1, QS2;

	TQuaternion QV(V2);
    QS1 = q20;
    QS1.MultiplyLeft(V2);
    QS2 = q20;
    QS2.MultiplyLeft(QV);
    cout << "Compare Q.multiplyLeft(TQuaternion) && Q.MultiplyLeft(TVector3) " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
    if (QS1!=q20) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
    TQuaternion q59 = V2 * q20;
    cout << "q59 = V2 * q20 " << endl;
    q59.Print();
    if (QS1==q59) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

    QS1 = q20;
    QS1*=V2;
    QS2 = q20;
    QS2*=QV;
    cout << "Compare Q*=TQuaternion && Q*=TVector3 " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

    QS1 = QV * q20;
    QS2 = V2 * q20;
    cout << "Compare QV * q20 && V2 * q20 " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

    QS1 = q20;
    QS1.DivideLeft(V2);
    QS2 = q20;
    QS2.DivideLeft(QV);
    cout << "Compare Q.DivideLeft(TQuaternion) && Q.DivideLeft(TVector3) " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

    QS1 = q20;
    QS1/=V2;
    QS2 = q20;
    QS2/=QV;
    cout << "Compare Q/=TQuaternion) && Q/=TVector3) " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

    QS1 = V2 / q20;
    QS2 = QV / q20;
    cout << "Compare V2 / q20 && QV / q20 " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

    QS1 = q20.LeftProduct(V2);
    QS2 = q20.LeftProduct(QV);
    cout << "Compare Q.LeftProduct(TQuaternion) && Q.LeftProduct(TVector3) " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

    QS1 = q20*V2;
    QS2 = q20*QV;
    cout << "Compare Q* TQuaternion && Q* TVector3 " << endl;
    QS1.Print();
    QS2.Print();

    QS1 = q20.LeftQuotient(V2);
    QS2 = q20.LeftQuotient(QV);
    cout << "Compare Q.LeftQuotient( TQuaternion ) && Q.LeftQuotient( TVector3 ) " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

    QS1 = q20 / V2;
    QS2 = q20 / QV;
    cout << "Compare Q / TQuaternion && Q / TVector3 " << endl;
    QS1.Print();
    QS2.Print();
    if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

  // ------------- quaternion algebra
   cout << endl << "============ TESTING quaternion algebra  " << endl;

   TQuaternion q30(0.3,0.4,0.5,0.6);
   TQuaternion q31(0.2,0.4,0.5,0.59);

   cout << "q30(0.3,0.4,0.5,0.6) " << endl;
   q30.Print();
   cout << "q31(0.2,0.4,0.5,0.59) " << endl;
   q31.Print();

   TQuaternion q32 = q30 + q31;
   cout << "q32 = q30 + q31 " << endl;
   q32.Print();
   q30+=q31;
   cout << "q30+=q31 " << endl;
   q30.Print();
   if (q30==q32) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   TQuaternion q33 = q32 - TQuaternion(0.1,0.5,0.6,0.7);
   cout << "q33 = q32 - TQuaternion(0.1,0.5,0.6,0.7) " << endl;
   q33.Print();
   q30 -= TQuaternion(0.1,0.5,0.6,0.7);
   cout << "q30-=TQuaternion(0.1,0.5,0.6,0.7)) " << endl;
   q30.Print();
   if (q30==q33) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   cout << " Norm q30 " << q30.Norm() << " Norm q31 " << q31.Norm() << " prod= "
                                                   << q30.Norm() * q31.Norm() << endl;
   TQuaternion q34 = q30 * q31;
   cout << "q34 = q30 * q31 " << endl;
   q34.Print();
   q30*=q31;
   cout << "q30*=q31 " << endl;
   q30.Print();
   cout << "norm = " << q30.Norm() << endl;
   if (q30==q34) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   TQuaternion q35 = q30 / q19;
   cout << "q35 = q30 / q19 " << endl;
   q35.Print();
   cout << " Norm q30 " << q30.Norm() << " Norm q19 " << q19.Norm() << " div= "
                                                   << q30.Norm() / q19.Norm() << endl;
   cout << " Norm q35 " << q35.Norm() << endl;
   q30/=q19;
   cout << "q30/=q19 " << endl;
   q30.Print();
   cout << "norm = " << q30.Norm() << endl;
   if (q30==q35) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   // compare left / right product
   cout << endl << "============ TESTING left / right product  " << endl;

   TQuaternion QL = q9;
   TQuaternion QR = q17;

   QS1 = QL;
   QS1 *= QR;
   QS2  = QR;
   QS2.MultiplyLeft(QL);
   cout << " compare QL *= QR && QR.MultiplyLeft(QL) " << endl;
   QS1.Print();
   QS2.Print();
   if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   QS1 = QL;
   QS1 /= QR;
   QS2 = QR;
   QS2.DivideLeft(QL);
   QS2 = QS2.Invert();
   cout << " compare QL /= QR && QR.DivideLeft(QL).Invert() " << endl;
   QS1.Print();
   QS2.Print();
   if (QS1==QS2) cout << "OK !" << endl;
   else {
		cout << " ! NOT OK ! " << endl;
		cout << "DIFF = " << endl;
		(QS1-QS2).Print();
		ratio = (QS1-QS2).Norm() / QS1.Norm();
		if (ratio < 1e-6) {
			cout << " => OK : ratio [(QL /= QR) - (QR.DivideLeft(QL).Invert()).Norm()] / [QL /= QR] "<< ratio << " is infinitesimal -> acceptable (double inversion cannot be numerically exact (1/(1/x)) != x) "<< endl;
		}
   }

   QS1 = QL * QR;
   QS2 = QR.LeftProduct(QL);
   cout << " compare QL * QR && QR.LeftProduct(QL) " << endl;
   QS1.Print();
   QS2.Print();
   if (QS1==QS2) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   QS1 = QL / QR;
   QS2 = QR.LeftQuotient(QL).Invert();
   cout << " compare QL / QR && QR.LeftQuotient(QL).Invert() " << endl;
   QS1.Print();
   QS2.Print();
   if (QS1==QS2) cout << "OK !" << endl;
   else {
	   cout << " ! NOT OK ! " << endl;
	   cout << "DIFF = " << endl;
		(QS1-QS2).Print();
		ratio = (QS1-QS2).Norm() / QS1.Norm();
		if (ratio < 1e-6) {
			cout << " => OK : ratio [(QL / QR) - (QR.LeftQuotient(QL).Invert()).Norm()] / [QL / QR] "<< ratio << " is infinitesimal -> acceptable (double inversion cannot be numerically exact (1/(1/x)) != x) "<< endl;
		}
   }

   // inversion
   cout << endl << "============ TESTING inversion  " << endl;
   QS1 *= QS1.Invert();
   cout << " QS1 *= QS1.Invert() " << endl;
   QS1.Print();
   cout << "norm " << QS1.Norm() << endl;

   cout << " QS2 " << endl;
   QS2.Print();
   QS2.Normalize();
   cout << " QS2.Normalize() " << endl;
   QS2.Print();
   cout << " Norm =" << QS2.Norm() << endl;
   QS1 = QS2.Conjugate();
   cout << " Conjugate , Norm " << QS1.Norm() << endl;
   QS1.Print();
   if (QS2.Norm() == QS1.Norm()) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   cout << endl << "============ TESTING SetQAngle / GetQAngle & Rotations  " << endl;
   QS1.SetQAngle(0.3);
   cout << " QS1.SetQAngle(0.3) -> angle  " <<  QS1.GetQAngle() << endl;
   if (QS1.GetQAngle() == 0.3) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;
   QS2 = QS1;
   QS2.Normalize();
   cout << " norm QS1  : " << QS1.Norm() << " norm QS2 = QS1.Normalize() " << QS2.Norm() << endl;
   cout << " QS2 = QS1 ; QS2.Normalize() -> angle  " <<  QS2.GetQAngle() << endl;
   if (QS2.GetQAngle() == 0.3) cout << "OK !" << endl; else cout << " ! NOT OK ! " << endl;

   cout << " Testing angle of rotation versus angle of quaternion " << endl;
   TVector3 v = QS1.Rotation(V2);
   cout << " V2 : " << endl;
   V2.Print();
   cout << " v : " << endl;
   v.Print();
   cout << " norm V2 = "<< V2.Mag() << " norm QS1.Rotation(V2) = " << v.Mag() << endl;
   TVector3 VPr2 = QS1.fVectorPart*((QS1.fVectorPart*V2)/QS1.fVectorPart.Mag2());
   TVector3 VPr = QS1.fVectorPart*((QS1.fVectorPart*v)/QS1.fVectorPart.Mag2());
   cout << " Projection V2 on rotation axis (QS1): " << endl;
   VPr2.Print();
	cout << " Projection QS1.Rotation(V2) on rotation axis (QS1) : " << endl;
   VPr.Print();
   if (VPr2.Mag() == VPr.Mag()) cout << "OK !" << endl;
   else {
      cout << " ! NOT OK ! " << endl;
      cout << "DIFF = " << endl;
      (VPr2-VPr).Print();
	  ratio = (VPr2-VPr).Mag()/VPr2.Mag();
		if (ratio < 1e-6) {
			cout << " => OK : ratio Norm[Projection V2 - Projection QS1.Rotation(V2) on rotation axis (QS1)] /  Norm(V2) "<< ratio << " is infinitesimal -> acceptable (double inversion cannot be numerically exact (1/(1/x)) != x) "<< endl;
		}
   }

   TVector3 VP2 = V2 - VPr2;
   TVector3 VP = v - VPr;
   cout << " Projection V2 perp. rotation axis : " << endl;
   VP2.Print();
   cout << " Projection QS1.Rotation(V2) perp. rotation axis : " << endl;
   VP.Print();
   cout << " norm Projection V2 perp. rotation axis = "<< VP2.Mag()
	   << " norm Projection QS1.Rotation(V2) perp. rotation axis = " << VP.Mag() << endl;
   if (VP2.Mag() == VP.Mag()) cout << "OK !";
   else {
	  cout << " ! NOT OK ! " << endl;
	  ratio = fabs(VP2.Mag()-VP.Mag())/VP2.Mag();
		if (ratio < 1e-6) {
			cout << " => OK : ratio Norm[Projection V2 - Projection QS1.Rotation(V2) perp. rotation axis ] /  Norm(V2) "<< ratio << " is infinitesimal -> acceptable "<< endl;
		}
   }
   double cosangle = (VP*VP2)/VP2.Mag2();
   double angle = acos(cosangle);
   cout << " cos( QS1.Rotation(V2) * V2/V2.Mag2()) = " << cosangle
      <<" arccos( QS1.Rotation(V2) * V2/V2.Mag2()) " << angle << endl;
   if (angle == QS1.GetQAngle()*2 ) cout << "OK !" << endl;
   else {
	   cout << " ! NOT OK ! " << endl;
 	  ratio = fabs(angle-QS1.GetQAngle()*2)/angle;
		if (ratio < 1e-6) {
			cout << " => OK : ratio abs(angle-QS1.GetQAngle()*2)/angle "<< ratio << " is infinitesimal -> acceptable "<< endl;
		}
  }

   QS1.Rotate(V2);
   cout << "compare Rotation  / Rotate " << endl;
   v.Print();
   V2.Print();

   TVector3 V4(0,0,1);
   cout << " V4(0,0,1) " << endl;
   V4.Print();
   TQuaternion q(cos(45*TMath::DegToRad()),sin(45*TMath::DegToRad()),0,0);
   q.Rotate(V4);
   cout << " (0,0,1) -> 90 rotation about X axis " << endl;
   V4.Print();
   q.SetAxisQAngle(TVector3(0,0,1),45*TMath::DegToRad());
   cout << " result -> 90 rotation about Z axis " << endl;
   q.Rotate(V4);
   V4.Print();
   q.SetAxisQAngle(TVector3(0,1,0),45*TMath::DegToRad());
   cout << " result -> 90 rotation about Y axis " << endl;
   q.Rotate(V4);
   V4.Print();
   q.SetAxisQAngle(TVector3(0,0,1),45*TMath::DegToRad());
   cout << " result -> 90 rotation about Z axis " << endl;
   q.Rotate(V4);
   V4.Print();

   // test infinitesimal
   TQuaternion q(cos(1e-6),sin(1e-3),0,0);
   cout << " V4 " << endl;
   V4.Print();
   q.Rotate(V4);
   cout << " v4 -> eps rotation about X axis " << endl;
   V4.Print();
   q.SetAxisQAngle(TVector3(0,0,1),1e-3);
   cout << " result -> eps rotation about Z axis " << endl;
   q.Rotate(V4);
   V4.Print();
   q.SetAxisQAngle(TVector3(0,1,0),1e-3);
   cout << " result -> eps rotation about Y axis " << endl;
   q.Rotate(V4);
   V4.Print();
   q.SetAxisQAngle(TVector3(0,0,1),1e-3);
   cout << " result -> eps rotation about Z axis " << endl;
   q.Rotate(V4);
   V4.Print();

}
